package edu.mit.ll.group43.surfaceoptimization.aircraft;

import java.io.FileNotFoundException;
import java.util.HashMap;

import edu.mit.ll.group43.surfaceoptimization.input.MinSeparationReader;
import edu.mit.ll.group43.surfaceoptimization.input.TextMinSeparationReader;

/**
 * This class reads in and returns the minimum separation times
 * for departures based on the weight classes of both the
 * leading and following aircraft. The required times are based
 * on the weight classes defined in DefaultWeightClass.
 * 
 * @author William Hawkins
 *
 */
public class DefaultDepartureMinimumSeparationCalculator implements MinimumSeparationCalculator<DefaultWeightClass> {
	
	public static DefaultDepartureMinimumSeparationCalculator instance;
	
	private final String SEP_FILE = "data/separations/ddsep";
	private MinSeparationReader sepReader;
	private HashMap<DefaultWeightClass, HashMap<DefaultWeightClass, Integer>> ddseparations;
	
	public static DefaultDepartureMinimumSeparationCalculator getInstance() throws FileNotFoundException {
		if (instance == null) {
			instance = new DefaultDepartureMinimumSeparationCalculator();
		}
		return instance;
	}
	
	/**
	 * Constructor.
	 * 
	 * @throws FileNotFoundException
	 */
	private DefaultDepartureMinimumSeparationCalculator() throws FileNotFoundException {
		initSeparations();
	}
	
	/**
	 * Reads in the minimum separation times from a file and stores them
	 * in the HashMap. Leading rows, trailing columns.
	 * 
	 * @throws FileNotFoundException
	 */
	private void initSeparations() throws FileNotFoundException {
		
		ddseparations = new HashMap<DefaultWeightClass, HashMap<DefaultWeightClass, Integer>>();
		for (DefaultWeightClass wc : DefaultWeightClass.values()) {
			ddseparations.put(wc, new HashMap<DefaultWeightClass, Integer>());
		}
		
		sepReader = new TextMinSeparationReader(SEP_FILE);
		
		int leading = 0;
		int[] times;
		
		while ((times = sepReader.nextLine()) != null) {
			DefaultWeightClass leadingwc = DefaultWeightClass.values()[leading];
			for (int i = 0; i < DefaultWeightClass.values().length; i ++) {
				ddseparations.get(leadingwc).put(DefaultWeightClass.values()[i], times[i]);
			}
			leading ++;
		}
	}
	
	/**
	 * Returns the required separation between the given leading and
	 * following aircraft. Departure actions are assumed.
	 * 
	 */
	@Override
	public int getSeparation(WeightClass leader, ActionType a1, WeightClass follower, ActionType a2) {
		return ddseparations.get(leader).get(follower);
	}
}
