package edu.mit.ll.group43.surfaceoptimization.dp;

import java.util.ArrayList;
import java.util.Date;
import java.util.TreeSet;

import edu.mit.ll.group43.surfaceoptimization.Driver;
import edu.mit.ll.group43.surfaceoptimization.Sequencer;
import edu.mit.ll.group43.surfaceoptimization.aircraft.Aircraft;
import edu.mit.ll.group43.surfaceoptimization.analysis.Logger;
import edu.mit.ll.group43.surfaceoptimization.dp.factory.DPSequencerFactory;
import edu.mit.ll.group43.surfaceoptimization.dp.factory.cost.CostStrategy;
import edu.mit.ll.group43.surfaceoptimization.dp.factory.netgen.CPSNetworkGenerator;
import edu.mit.ll.group43.surfaceoptimization.dp.factory.sequencerecovery.SequenceRecoveryStrategy;
import edu.mit.ll.group43.surfaceoptimization.network.NetworkEdge;
import edu.mit.ll.group43.surfaceoptimization.network.dp.CPSNetworkNode;
import edu.mit.ll.group43.surfaceoptimization.network.dp.SubsequenceGeneratorNode;

/**
 * Optimizes a sequence using Dynamic Programming, with respect to minimum
 * separation times and Constrained Position Shifting (CPS). The algorithm
 * used to do this can be altered through the use of an object whose class
 * implements the DPSequencerFactory interface.
 * 
 * @author William Hawkins
 *
 */
public class DPSequencer extends Sequencer {
	
	private CostStrategy costStrategy;
	private SequenceRecoveryStrategy sequenceRecoveryStrategy;
	private CPSNetworkGenerator cpsNetworkGenerator;
	
	/**
	 * Constructor
	 * 
	 * @param cps
	 */
	public DPSequencer(DPSequencerFactory factory) {
		super(factory.getCPS(), factory.getCalc());
		
		this.costStrategy = factory.getCostStrategy();
		this.sequenceRecoveryStrategy = factory.getSequenceRecoveryStrategy();
		this.cpsNetworkGenerator = factory.getCPSNetworkGenerator();
	}
	
	/**
	 * Test method. Prints out the generated nodes.
	 * 
	 * @param stages
	 */
	public void printGeneratedNodes(ArrayList<ArrayList<SubsequenceGeneratorNode>> stages) {
		Driver.debug.println("Generator Network:");
		int stageNum = 0;
		for (ArrayList<SubsequenceGeneratorNode> stage : stages) {
			stageNum ++;
			Driver.debug.println("==========");
			Driver.debug.println("Stage " + stageNum);
			Driver.debug.println("----------");
			for (SubsequenceGeneratorNode node : stage) {
				for (ArrayList<Integer> path : node.getValidPaths()) {
					String spath = "";
					for (Integer i : path) {
						spath += i + " ";
					}
					Driver.debug.println(spath);
				}
			}
		}
	}
	
	/**
	 * Test method. Prints generated CPS nodes.
	 * @param stages
	 */
	public void printCPSNodes(ArrayList<TreeSet<CPSNetworkNode>> stages) {
		Driver.debug.println("CPS Network:");
		int stageNum = 0;
		for (TreeSet<CPSNetworkNode> stage : stages) {
			stageNum ++;
			Driver.debug.println("==========");
			Driver.debug.println("Stage " + stageNum);
			Driver.debug.println("----------");
			for (CPSNetworkNode node : stage) {
				Driver.debug.println(node.getAircraft().toString());
			}
		}
	}
	
	public ArrayList<TreeSet<CPSNetworkNode>> removeNextNodes(ArrayList<TreeSet<CPSNetworkNode>> network) {
		for (TreeSet<CPSNetworkNode> stage : network) {
			for (CPSNetworkNode node : stage) {
				node.setNextNodes(null);
			}
		}
		return network;
	}
	
	/**
	 * Optimally re-sequences the given FCFS sequence with respect to
	 * minimum separation times and CPS using Dynamic Programming.
	 * 
	 */
	@Override
	public ArrayList<Aircraft> sequence(ArrayList<Aircraft> seq) {
		
		ArrayList<TreeSet<CPSNetworkNode>> network = 
				cpsNetworkGenerator.generateNetwork(seq);
		
		int arcs = network.get(0).size();
		for (int i = 0; i < network.size()-1; i ++) {
			for (CPSNetworkNode node : network.get(i)) {
				arcs += node.getNextNodes().size();
			}
		}
		arcs += network.get(network.size()-1).size();
		if (Logger.logging) {
			Driver.debug.println("Arcs created in network: " + arcs);
		}
		
		// Remove all nextNodes to save memory; they are no longer needed
		network = removeNextNodes(network);
		
		Date start = new Date();
		// Initialize the first stage
		costStrategy.initializeFirstStage(network);
		
		// Solve the remaining stages
		for (int i = 1; i < network.size(); i ++) {
			for (CPSNetworkNode node : network.get(i)) {
				if (Logger.logging) {
					Logger.predNodes(node.getPrevNodes().size());
				}
				NetworkEdge optPrevEdge = node.getPrevNodes().get(0);
				CPSNetworkNode optPrevNode = (CPSNetworkNode) optPrevEdge.getSource();
				int shortest = optPrevNode.getOptDistance() + costStrategy.getEdgeWeight(optPrevEdge);
				
				// Logger values
				boolean tie = false;
				int ties = 0;
				
				for (int j = 1; j < node.getPrevNodes().size(); j ++) {
					NetworkEdge prevEdge = node.getPrevNodes().get(j);
					CPSNetworkNode prevNode = (CPSNetworkNode) prevEdge.getSource();
					int distance = prevNode.getOptDistance() + costStrategy.getEdgeWeight(prevEdge);
					
					if (distance < shortest) {
						shortest = distance;
						optPrevEdge = prevEdge;
					} else if (Logger.logging && distance == shortest) {
						Logger.tie(i+1);
						tie = true;
						ties ++;
					}
				}
				
				if (Logger.logging && tie) {
					Logger.tieGroup(ties + 1);
				}
				
				costStrategy.setOptimalValues(optPrevEdge, shortest);
				sequenceRecoveryStrategy.processPrevNodes(node);
			}
			// The last stage is no longer needed, so get rid of it to
			// save memory
			network.set(i-1, null);
		}
		
		// Find the last node of the optimal path
		CPSNetworkNode optEnd = network.get(network.size()-1).first();
		int shortest = optEnd.getOptDistance();
		for (CPSNetworkNode node : network.get(network.size()-1)) {
			int distance = node.getOptDistance();
			if (distance < shortest) {
				optEnd = node;
				shortest = distance;
			}
		}
		
		// Recover and return the optimal sequence from that node
		ArrayList<Aircraft> optimalSequence = sequenceRecoveryStrategy.recoverSequence(optEnd);
		Date end = new Date();
		if (Logger.timing) {
			Logger.solveNetwork(end.getTime()-start.getTime());
		}
		return optimalSequence;
	}
}