package edu.mit.ll.group43.surfaceoptimization.dp.factory.cost;

import java.util.ArrayList;
import java.util.Collections;

import edu.mit.ll.group43.surfaceoptimization.aircraft.Aircraft;
import edu.mit.ll.group43.surfaceoptimization.aircraft.AircraftAlphaComparator;
import edu.mit.ll.group43.surfaceoptimization.aircraft.MinimumSeparationCalculator;
import edu.mit.ll.group43.surfaceoptimization.aircraft.MinimumSeparationCalculator.ActionType;
import edu.mit.ll.group43.surfaceoptimization.aircraft.WeightClass;
import edu.mit.ll.group43.surfaceoptimization.network.NetworkEdge;
import edu.mit.ll.group43.surfaceoptimization.network.dp.CPSNetworkNode;

/**
 * The spot cost strategy for the One Gap algorithm, which compares
 * scheduling the current aircraft to take off between the last two
 * projected takeoffs, if it can fit, against scheduling it to
 * take off last.
 * 
 * @author William Hawkins
 *
 */
public class OneGapSpotCostStrategy extends SpotCostStrategy {

	public OneGapSpotCostStrategy(int cps,
			MinimumSeparationCalculator<? extends WeightClass> calc) {
		super(cps, calc);
	}
	
	/**
	 * Returns a list of aircraft sorted in order of when they are
	 * predicted to take off, based on their optimal spot release
	 * times.
	 * 
	 * @param edge
	 * @return
	 */
	public ArrayList<Aircraft> sortedAlphasSoFar(NetworkEdge edge) {
		ArrayList<Aircraft> alphaSort = new ArrayList<Aircraft>();
		CPSNetworkNode optPrevNode = (CPSNetworkNode) edge.getSource();
		alphaSort.add(optPrevNode.getLastAircraft());
		
		NetworkEdge prevEdge = edge;
		while ((prevEdge = ((CPSNetworkNode) prevEdge.getSource()).getOptPrevNode()) != null) {
			alphaSort.add(((CPSNetworkNode)prevEdge.getSource()).getLastAircraft());
		}
		Collections.sort(alphaSort, AircraftAlphaComparator.getInstance());
		
		return alphaSort;
	}
	
	/**
	 * Determines if lastReleased can take off before lastTakeOff, given that it
	 * can leave the spot no earlier than earlyReleaseBound, without increasing
	 * the spot delay of lastTakeOff or incurring any delay at the runway.
	 * 
	 * @param earlyReleaseBound
	 * @param lastReleased
	 * @param lastTakeOff
	 * @return
	 */
	public boolean canFitBefore(int earlyReleaseBound, Aircraft lastReleased, Aircraft lastTakeOff) {
		int lateSeparation = CALC.getSeparation(lastReleased.getWeightClass(), ActionType.DEPARTURE,
				lastTakeOff.getWeightClass(), ActionType.DEPARTURE);
		return earlyReleaseBound + lastReleased.getUnimpTaxiTime() + lateSeparation
				<= lastTakeOff.getRunwayReadyTime();
	}
	
	/**
	 * Determines the earliest time at which follower can be released so that it
	 * takes off after leader without incurring any runway delay, with respect
	 * to the given predecessor bound.
	 * 
	 * @param predBound
	 * @param leader
	 * @param follower
	 * @return
	 */
	public int earliestReleaseAfter(int predBound, Aircraft leader, Aircraft follower) {
		int earlyReleaseBound = Math.max(follower.getSpotReadyTime(), predBound);
		
		// If there is a first aircraft (i.e. b isn't the first one in the
		// sequence), then set the separation between a and the current
		// aircraft and update the early release bound
		if (leader != null) {
			int earlySeparation = CALC.getSeparation(leader.getWeightClass(), ActionType.DEPARTURE,
					follower.getWeightClass(), ActionType.DEPARTURE);
			earlyReleaseBound = Math.max(earlyReleaseBound,
					leader.getRunwayReadyTime() + earlySeparation - follower.getUnimpTaxiTime());
		}
		
		return earlyReleaseBound;
	}
	
	/**
	 * Determines the earliest time at which follower can be released so that it
	 * takes off after leader without incurring any runway delay, with respect
	 * to the predecessor bound according to the given previous node.
	 * 
	 * @param optPrevNode
	 * @param leader
	 * @param follower
	 * @return
	 */
	public int earliestReleaseAfter(CPSNetworkNode optPrevNode, Aircraft leader, Aircraft follower) {
		return earliestReleaseAfter(optPrevNode.getLastAircraft().getOptReleaseTime(),
				leader, follower);
	}
	
	/**
	 * Determines the earliest time at which lastReleased can be released from the
	 * spot to take off between aircraft a and b without increasing the spot delay
	 * of b or incurring any delay at the runway. Returns -1 if lastReleased cannot
	 * be fit between a and b while following these constraints.
	 * 
	 * @param optPrevNode Optimal previous node in the network for lastReleased.
	 * @param a
	 * @param b
	 * @param lastReleased
	 * @return
	 */
	public int earliestReleaseBetween(CPSNetworkNode optPrevNode, Aircraft a, Aircraft b, Aircraft lastReleased) {
		return earliestReleaseBetween(optPrevNode.getLastAircraft().getOptReleaseTime(), a, b, lastReleased);
	}
	
	/**
	 * Determines the earliest time at which lastReleased can be released from the
	 * spot to take off between aircraft a and b without increasing the spot delay
	 * of b or incurring any delay at the runway. Returns -1 if lastReleased cannot
	 * be fit between a and b while following these constraints.
	 * 
	 * @param predBound
	 * @param a
	 * @param b
	 * @param lastReleased
	 * @return
	 */
	public int earliestReleaseBetween(int predBound, Aircraft a, Aircraft b, Aircraft lastReleased) {
		int earlyReleaseBound = earliestReleaseAfter(predBound, a, lastReleased);
		
		// If this aircraft can fit between a and b (release time + taxi time + separation
		// <= b's arrival at the runway) then return that release time
		if (canFitBefore(earlyReleaseBound, lastReleased, b)) {
			return earlyReleaseBound;
		}
		
		// Otherwise, it can't be done
		return -1;
	}
	
	/**
	 * Determines the earliest time at which the last aircraft of the target node
	 * of the given network edge can be released from the spot. Considers the gap
	 * between the last two takeoffs in the current projected takeoff sequence and
	 * tries to fit the current aircraft into it. Failing that, it sends it as early
	 * as it can to minimize how long it takes off after the last projected takeoff
	 * time, while respecting the optimal spot release time of this aircraft's
	 * direct predecessor in the CPS spot release sequence.
	 * 
	 * @param edge
	 */
	@Override
	public int earliestRelease(NetworkEdge edge) {
		
		Aircraft lastReleasedAircraft = ((CPSNetworkNode) edge.getTarget()).getLastAircraft();
		CPSNetworkNode optPrevNode = (CPSNetworkNode) edge.getSource();
		
		// Sort all previous aircraft by the order in which they will reach
		// the runway (and take off)
		ArrayList<Aircraft> alphaSort = sortedAlphasSoFar(edge);
		
		// Get the last two aircraft to take off according to the given optimal
		// release schedule
		Aircraft firstReadyAircraft = null;
		if (alphaSort.size() >= 2) {
			firstReadyAircraft = alphaSort.get(alphaSort.size()-2);
		}
		Aircraft secondReadyAircraft = alphaSort.get(alphaSort.size()-1);
		
		// Determine the earliest that this aircraft can be released while satisfying
		// all constraints
		int earlyRelease = earliestReleaseBetween(optPrevNode, firstReadyAircraft, 
				secondReadyAircraft, lastReleasedAircraft);
		if (earlyRelease > -1) {
			return earlyRelease;
		}
		
		// Find out the earliest it can be sent after the last aircraft predicted to
		// take off, while still respecting this aircraft's SRS predecessor
		return earliestRelease(secondReadyAircraft, lastReleasedAircraft,
				optPrevNode.getLastAircraft().getOptReleaseTime());
	}

}
