package edu.mit.ll.group43.surfaceoptimization.dp.factory.cost;

import java.util.ArrayList;
import java.util.TreeSet;

import edu.mit.ll.group43.surfaceoptimization.aircraft.Aircraft;
import edu.mit.ll.group43.surfaceoptimization.aircraft.MinimumSeparationCalculator;
import edu.mit.ll.group43.surfaceoptimization.aircraft.MinimumSeparationCalculator.ActionType;
import edu.mit.ll.group43.surfaceoptimization.aircraft.WeightClass;
import edu.mit.ll.group43.surfaceoptimization.network.NetworkEdge;
import edu.mit.ll.group43.surfaceoptimization.network.dp.CPSNetworkNode;

/**
 * Cost strategy for optimizing at the spot by setting the
 * optimal spot release times, in addition to the total
 * distance from the start and the optimal previous node.
 * 
 * @author William Hawkins
 *
 */
public class SpotCostStrategy extends CostStrategy {

	public SpotCostStrategy(int cps,
			MinimumSeparationCalculator<? extends WeightClass> calc) {
		super(cps, calc);
	}

	/**
	 * Initializes all nodes in the first stage of the network
	 * to have a distance of zero from the beginning, no
	 * previous node, and for their last aircraft to have an
	 * optimal spot release time equal to that of their spot
	 * ready time.
	 * 
	 * @param network
	 */
	@Override
	public void initializeFirstStage(ArrayList<TreeSet<CPSNetworkNode>> network) {
		for (CPSNetworkNode node : network.get(0)) {
			node.setOptDistance(0);
			node.setOptPrevNode(null);
			
			Aircraft last = node.getLastAircraft();
			last.setOptReleaseTime(last.getSpotReadyTime());
		}
	}

	/**
	 * Sets the optimal distance from the beginning, the optimal
	 * previous node, and the optimal spot release time of the
	 * last aircraft in the target node.
	 * 
	 * @param edge Edge whose target will have its last aircraft
	 * 				solved for
	 * @param distance Total amount of delay incurred so far
	 */
	@Override
	public void setOptimalValues(NetworkEdge edge, int distance) {
		CPSNetworkNode prev = (CPSNetworkNode) edge.getSource();
		CPSNetworkNode node = (CPSNetworkNode) edge.getTarget();
		Aircraft follower = node.getLastAircraft();
		
		node.setOptDistance(distance);
		node.setOptPrevNode(edge);
		follower.setOptReleaseTime(distance - prev.getOptDistance() + follower.getSpotReadyTime());
	}
	
	/**
	 * Returns the earliest time that the follower aircraft can leave
	 * at with the leader aircraft as its direct predecessor, while
	 * obeying all constraints.
	 * 
	 * @param leader
	 * @param follower
	 * @return
	 */
	public int earliestRelease(Aircraft leader, Aircraft follower) {
		return earliestRelease(leader, follower, leader.getOptReleaseTime());
	}
	
	/**
	 * Returns the earliest time that the follower aircraft can leave
	 * at with the leader aircraft as its direct predecessor, while
	 * obeying all constraints.
	 * 
	 * @param leader
	 * @param follower
	 * @param predBound The optimal spot leave time of follower's direct
	 * 				predecessor in the spot release sequence.
	 * @return
	 */
	public int earliestRelease(Aircraft leader, Aircraft follower, int predBound) {
		int separation = CALC.getSeparation(leader.getWeightClass(), ActionType.DEPARTURE,
				follower.getWeightClass(), ActionType.DEPARTURE);
		
		// Calculate the best starting time for the follower, given
		// the leader as its immediate predecessor in the optimal
		// spot release sequence
		int bestStart = separation-(follower.getUnimpTaxiTime()-leader.getUnimpTaxiTime())+leader.getOptReleaseTime();
		bestStart = Math.max(bestStart, follower.getSpotReadyTime());
		bestStart = Math.max(bestStart, predBound);
		
		return bestStart;
	}
	
	/**
	 * Returns the earliest release time for the last
	 * aircraft of the target without incurring any
	 * delay at the runway, without leaving until it
	 * is ready and without leaving before the source's
	 * last aircraft has already been released.
	 * 
	 * @param edge
	 * @return
	 */
	public int earliestRelease(NetworkEdge edge) {
		Aircraft leader = ((CPSNetworkNode) edge.getSource()).getLastAircraft();
		Aircraft follower = ((CPSNetworkNode) edge.getTarget()).getLastAircraft();
		return earliestRelease(leader, follower);
	}
	
	/**
	 * Returns the calculated edge weight, based on what has
	 * already been solved for the source node in the given
	 * network edge.
	 * 
	 * @param edge
	 */
	@Override
	public int getEdgeWeight(NetworkEdge edge) {
		
		Aircraft follower = ((CPSNetworkNode) edge.getTarget()).getLastAircraft();
		int bestStart = earliestRelease(edge);
		
		// Return the delay that is added to the system delay by
		// releasing the follower at time 'bestStart'
		int addedDelay = bestStart - follower.getSpotReadyTime();
		return addedDelay;
	}
}
